#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "compare.h"
#include "pasv.h"
#include "misc.h"
#include "io.h"
#include "fnmatch.h"
#include "log.h"
#include "date.h"


#define MAX_FIRST_FILES 10

static int num_first_files = 0;
static char *first_files[ MAX_FIRST_FILES ];


#define MAX_IGNORE_FILES 20

static int num_ignore_files = 0;
static char *ignore_files[ MAX_IGNORE_FILES ];












//
// This function takes to two linked lists of top directory entries
// from both sites, mirror->A_host->current_dir and B_host respectively.
//
// Tries to work out which files and directories should be move to the other
// directory by building the initial list
//
// Into ->mirror_paths
//
void compare_dirs(mirror_t *mirror, dir_t **target)
{
  dir_t *runner, *newd;


  // For now work out which dirs don't exist in the other dir and do those

  for (runner = mirror->A_host.current_dir;
       runner;
       runner = runner->next) {

    
    if (compare_test(runner, &mirror->B_host, mirror)) {  // Mirror this one.
      

      // Add runner to A_host->mirror_dirs
      newd = (dir_t *) malloc(sizeof(dir_t));
      if (!newd) {
	perror("malloc");
	return;
      }

      memset(newd, 0, sizeof(dir_t));

      newd->name = mystrcpy(runner->name);
      newd->size = runner->size;
      newd->type = runner->type;
      newd->date = runner->date;
      if (runner->dest_name)
	newd->dest_name = mystrcpy(runner->dest_name);


      newd->parent = mirror->A_host.mirror_current;


      pasv_addsorted(target, newd);


      log_printf(LOG_DEBUG, "Mirror A_host %s/%s/%s to B_host %s/%s/\n",
	     mirror->A_host.name, mirror->A_host.current_cwd, runner->name,
	     mirror->B_host.name, 
		 runner->dest_name ? runner->dest_name :
		 mirror->B_host.current_cwd);


    }


  }



  // FIXME!!
  // Free current_dir here

  pasv_freedir(&(mirror->A_host.current_dir));
  pasv_freedir(&(mirror->B_host.current_dir));

  mirror->A_host.current_dir = NULL;
  mirror->B_host.current_dir = NULL;


}





int compare_test(dir_t *dir, site_t *host, mirror_t *mirror)
{
  dir_t *runner;
  int   manual;
  static int last_manual = 0;

  if (mirror_verbose)
    log_printf(LOG_DEBUG, "[compare] thinking about '%s'...\n",
	       dir->name);



  // Check if it was manually listed
  // 1 it was
  if ((manual = mirror_ismanual(dir))) {

    last_manual = 1;
    mirror->site_max_age = 0;

    if (mirror_verbose) 
      // Set the age back so subdirectories get processed too.
      //mirror->max_file_age = 0;
      //mirror->max_dir_age  = 0;
      

      log_printf(LOG_DEBUG, "[compare] manual entry -- ignoring date checks: Max_age set to 0.\n",
		 dir->name);

    
  } else { // It wasn't manual this time, was it manual last run?


    if (last_manual) {
      // We are not manual now, but last go was, so reload max_age

      last_manual = 0;

      printf("Reloading date information\n");
      date_load(mirror, &(mirror->A_host));
      
    }
  }




  if (!manual && compare_ignore(dir)) {
    if (mirror_verbose) log_printf(LOG_DEBUG, "[compare] rejected. In ignore list\n");
    return 0;
  }


  if (dir->type == DT_FILE) {

    
    if ((mirror->move_empty_files == YN_NO) &&
	!dir->size) {

    if (mirror_verbose) log_printf(LOG_DEBUG, "[compare] rejected. 0 length file, conf=no\n");
      return 0;
    }
    // Check date

    
    if (mirror->max_file_age && (dir->date < (io_global_time - mirror->max_file_age))) {
      
      if (mirror_verbose) log_printf(LOG_DEBUG, "[compare] rejected. Older than max_file_age (%lu < %lu)\n",
				     dir->date, 
				     (io_global_time - mirror->max_file_age));
      
      return 0; // too old
    }

    

  } else { // directory


#if 0
    log_printf(LOG_DEBUG, "%s date %lu < %lu age\n", dir->name, dir->date,
	       mirror->site_max_age);
#endif


    if (!manual && mirror->site_max_age &&
	(dir->date <= mirror->site_max_age)) {

      if (mirror_verbose) log_printf(LOG_DEBUG, "[compare] rejected. Older than saved stamp (%lu < %lu)\n",
				     dir->date, 
				     mirror->site_max_age);

      return 0; // too old
    }

    if (!manual && mirror->max_dir_age && (dir->date < (io_global_time - mirror->max_dir_age))) {

      if (mirror_verbose) log_printf(LOG_DEBUG, "[compare] rejected. Older than max_dir_age (%lu < %lu)\n",
				     dir->date, 
				     (io_global_time - mirror->max_dir_age));

      return 0; // too old
    }

  }


  for (runner = host->current_dir;
       runner;
       runner = runner->next) {


    if (!compare_names(runner->name, dir->name)) {

      log_printf(LOG_DEBUG, "[compare] '%s' is the same as '%s'\n", runner->name, dir->name);

      // Right exists on both, if it is a file, check if it is shorter
      if ((dir->type == DT_FILE) && (runner->size < dir->size)) 
	return 1;

      // If it is a directory, check it.
      if (dir->type == DT_DIRECTORY) {

	// We need to save the destination name here.
	dir->dest_name = mystrcpy( runner->name );

	return 1;
      }

      return 0;
    }


  }


  return 1;

}





int compare_first(dir_t *node)
{
  int i;

  for (i = 0; i < num_first_files; i++)
    if (!fnmatch(first_files[i], node->name, FNM_CASEFOLD))
      return 1;

  return 0;

}



int compare_ignore(dir_t *node)
{
  int i;

  for (i = 0; i < num_ignore_files; i++)
    if (!fnmatch(ignore_files[i], node->name, FNM_CASEFOLD))
      return 1;

  return 0;

}



int compare_names(char *name1, char *name2)
{
  char *copy1, *copy2;



  copy1 = mystrcpy(name1);
  copy2 = mystrcpy(name2);

  if (!copy1 || !copy2)
    return 1;


  // replace all punctuations with *'s :)
  compare_fudge(copy1);
  compare_fudge(copy2);



#if 0
  printf("Matching '%s' vs '%s'.\n", copy1, copy2);
  if (!fnmatch(copy1, copy2, FNM_CASEFOLD)) {

    printf("copy1 = copy2\n");
    free(copy1);
    free(copy2);
    return 0;

  }
#endif 

  //  printf("Matching '%s' vs '%s'.\n", copy1, name2);
  if (!fnmatch(copy1, name2, FNM_CASEFOLD)) {

    free(copy1);
    free(copy2);
    return 0;
  }

  
  //  printf("Matching '%s' vs '%s'.\n", copy2, name1);
  if (!fnmatch(copy2, name1, FNM_CASEFOLD)) {
    
    free(copy1);
    free(copy2);
    return 0;
    
  }

  free(copy1);
  free(copy2);
  return 1;

}



void compare_fudge(char *name)
{
  char *r;


  while ((r = strpbrk(name, "._ -+=,!%()[]{}#~@?"))) {

    *r = '*';

  }

}




void compare_add2first(char *pattern)
{

  if (num_first_files >= MAX_FIRST_FILES) {

    log_printf(LOG_ERROR, "Warning! Too many first_files directives. Increase MAX_FIRST_FILES.\n");
    return;
  }


  first_files[ num_first_files++ ] = mystrcpy( pattern );

}


void compare_add2ignore(char *pattern)
{

  if (num_ignore_files >= MAX_IGNORE_FILES) {

    log_printf(LOG_ERROR, "Warning! Too many ignore_files directives. Increase MAX_IGNORE_FILES.\n");
    return;
  }


  ignore_files[ num_ignore_files++ ] = mystrcpy( pattern );

}
