#include <stdio.h>
#include <sys/types.h>
#include <errno.h>
#ifdef WIN32
#define WINDOWS_LEAN_AND_MEAN
#include <winsock.h>
#undef SO_ERROR   // blast, I clash with WIN32's error code :(:(:(
#else
#include <sys/time.h> /* big sigh to linux */
#include <unistd.h>
#endif



#include "mirror.h"
#include "io.h"
#include "state.h"
#include "sockets.h"
#include "site.h"
#include "pasv.h"
#include "cps.h"
#include "log.h"



time_t io_global_time = NULL;

static time_t time_last_check = NULL;

static int main_exit = 0;


// The fd set holders for select.
static fd_set io_fdset_read;
static fd_set io_fdset_write;


// The time struct for select's timeout
static struct timeval timeout;



void io_loop(mirror_t *mirror)
{

  int dsize;   // result from getdtablesize()
  int nret;    // return code from select()


  // Setup the timeval for timeout.
  // This is tuneable, but 30secs is pretty good for timeouts.
  timeout.tv_usec = IO_TIMEOUT_USEC;
  timeout.tv_sec  = IO_TIMEOUT_SEC;


  // setup dsize, on Win32 it is ignored, on
  // unix it retrieved from getdtablesize(), or
  // OPEN_MAX, or use POSIX getrlimit(), or ....
#ifndef WIN32
  
  dsize = getdtablesize();

#else

  dsize = 256;  // value is irrelevant in WIN32

#endif


  // Spin forever
  while (!main_exit) {

    // Clear the fdset's
    FD_ZERO( &io_fdset_read  );
    FD_ZERO( &io_fdset_write );


    time(&io_global_time);


    // call the state engine

    state_process( mirror, &mirror->A_host );
#ifndef ONE_SITE
    state_process( mirror, &mirror->B_host );
#endif


    if ((io_global_time - time_last_check) >= CPS_UPDATE) {

      time_last_check = io_global_time;

      cps_display( mirror );

    }





    // Add all fd's that are required

    //_addfd( &io_fdset_read, &io_fdset_write );

    timeout.tv_usec = IO_TIMEOUT_USEC;
    timeout.tv_sec  = IO_TIMEOUT_SEC;


    // Simply call select
  
    nret = select( dsize, &io_fdset_read, &io_fdset_write, NULL, &timeout);

    switch ( nret ) {

    case -1:    // error
      // The only annoying thing here is SunOS/HPUX and some Ultrix
      // can (frequently) return EAGAIN so we ignore it, and restart,
      // all other errors mean exit.

      if (( errno == EAGAIN ) || ( errno == EINTR ))
        continue;

      perror("io_loop: select()");

#ifdef WIN32
      printf("Windows error %d\n", WSAGetLastError());
#endif
      break;

      return;


    case 0:     // timeout - perform various timeout checks.
#if 0
      if (mirror->A_host.state.pasv == SO_ACTIVE) {
	char buffer[1];

	printf("ISSET %d and %d\n", mirror->A_host.pasv_socket, 
	       FD_ISSET(mirror->A_host.pasv_socket, &io_fdset_read));
	fflush(stdout);
	printf("read says %d errno %d\n",
	       read(mirror->A_host.pasv_socket, buffer, 1), errno);
	printf("2read says %d errno %d\n",
	       read(mirror->A_host.socket, buffer, 1), errno);

	printf("write %d\n", write(mirror->A_host.pasv_socket, buffer, 1));

      }
#endif

      break;


    default:    // there was something happening on our sockets.

      io_process(&mirror->A_host);
#ifndef ONE_SITE
      io_process(&mirror->B_host);
#endif

    } // switch

  } // while (1)


  // NOT REACHED




}



void io_set_read( int socket )
{

  FD_SET(socket, &io_fdset_read);

}


void io_set_write( int socket )
{

  FD_SET(socket, &io_fdset_write);

}


void io_process( site_t *node )
{

  switch(node->state.socket) {

  case SO_PENDING:
  case SO_LISTEN:


    if (FD_ISSET( node->socket, &io_fdset_write )) {
      
      node->state.socket = SO_ACTIVE;
      break;
      
    }
    
    if (FD_ISSET( node->socket, &io_fdset_read )) {
      
      sockets_read( node->socket, node->inbuffer, 1 );
      perror("connect()");
      log_printf(LOG_ERROR, "Failed to connect to %s\n", node->name);
      
      node->state.socket = SO_ERROR;

      state_set(node, ST_RETRY);
      
      break;

    }
    
    
    
    break; // ST_PENDING & ST_LISTEN
    
    
    
  case SO_ACTIVE:
    if (FD_ISSET( node->socket, &io_fdset_read )) {
      
      site_input(node);
      
    }
    break;


  case SO_BUFFERFULL:

      if (FD_ISSET( node->socket, &io_fdset_write )) {
      
        // ahh there is room to transmit again, purge the buffer
        site_output( node, NULL, 0);

      }
      
      break; // ST_BUFFERFULL


  default:
    ;
  }












  switch(node->state.pasv) {
    int new_socket;

  case SO_LISTEN:

    if (FD_ISSET( node->pasv_socket, &io_fdset_read )) {
      

      log_printf(LOG_DEBUG, "[%s] Calling accept()\n", node->name);


      new_socket = sockets_accept( node->pasv_socket, NULL, NULL );

      sockets_close( node->pasv_socket );
      node->pasv_socket = -1;


      if (new_socket == -1) {
	node->state.pasv = SO_ERROR;
	perror("accept");
	log_printf(LOG_ERROR, "Failed to accept connect from %s\n", node->name);
	break;
      }

      node->pasv_socket = new_socket;
      node->state.pasv = SO_PENDING;

      break;
      
    }
    break;


  case SO_PENDING:
    
    
    if (FD_ISSET( node->pasv_socket, &io_fdset_write )) {
      
      log_printf(LOG_DEBUG, "[%s] Connected SO_ACTIVE\n", node->name);

      node->state.pasv = SO_ACTIVE;
      break;
      
    }
    
    if (FD_ISSET( node->pasv_socket, &io_fdset_read )) {
      
      sockets_read( node->pasv_socket, node->pasvbuffer, 1 );
      perror("connect()");
      log_printf(LOG_ERROR, "Failed to passive connect to %s\n", node->name);
      
      node->state.pasv = SO_ERROR;
      
      break;
      
    }
    
    break; // ST_PENDING & ST_LISTEN
    
    
    
  case SO_ACTIVE:

    //    printf("checking pasv %d\n", node->pasv_socket);

    if (FD_ISSET( node->pasv_socket, &io_fdset_read )) {
      
      pasv_input(node);
      
    }
    break;
        
  default:
    ;
  }
  

}









void io_setexit(void)
{
  main_exit = 1;
}










