#include <stdio.h>
#include <stdlib.h>       // malloc
#include <string.h>       // memset
#if __STDC__ || WIN32
#include <stdarg.h>
#else
#include <varargs.h>
#endif
#include <ctype.h>

#include "site.h"
#include "mirror.h"
#include "sockets.h"
#include "state.h"
#include "misc.h"
#include "io.h"
#include "log.h"


int site_input(site_t *node)
{
  int result;
  char *line;   // Pointer to the current start of a string.


  // While we can get a line from the buffer parse it.
  // notice this can get seperated from here, so that we
  // simply call read() and fill the buffer here, and call
  // io_getline from elsewere. But not really required.

  // Reset the current position counter here.
  line = NULL;

  while((result = site_getline(node, &line)) > 0) { 

    // Pass the string on to next layer
    site_command(node, line);

  }

  // We are done.
  // We return 0 when there are no more info to process
  // and -1 signifies the connection was lost.

  return result;


}





int site_getline(site_t *node, char **last)
{ 
  int len;                      // return int of read().
  int left;                     // remaining items
  static char *next;            // strtok style PC holder. 



  // If last is NULL, it's the first call to this function for this
  // current iteration, so first we call read() to receive more data.
  if (!*last) {


    // Set it to start of the buffer
    *last = node->inbuffer;
    next = *last;


    // Retrieve more data from the socket.
    // We write it after any data that may already be left in the buffer
    // from a previous read. Generally doesn't happen, but we need to
    // allocate for it. (Buffer over/under -run hacking etc)

    if ((len = sockets_read(node->socket, &node->inbuffer[node->in_buffer], 
                            DATA_BUFSZ - node->in_buffer)) <= 0) {


      if (len == 0) {

	log_printf(LOG_DEBUG, "Connection to %s closed\n", node->name);
	
	state_set(node, ST_RETRY);

	return -1;

      }

      perror("read");
      
      log_printf(LOG_ERROR, "Connection to %s closed/lost (%d)\n", node->name, len);

      state_set(node, ST_RETRY);

      return -1;
    }


    node->idle = io_global_time;


    // Move along the PC according to how much was read.
    node->in_buffer += len;

    node->stats.bytes_from += len;


  } else {  // last == NULL

    // last isn't NULL, so we are now still parsing buffer content, if any

    *last = next;

  }


  // Right we now have some data in the buffer, extract a string.

  do {


    if ((int)(next - (char *)node->inbuffer) >= node->in_buffer) {
      

      // Compute what is remaining
      left = (node->in_buffer - (*last - (char *)node->inbuffer));


      // we only bother keeping the remaining if it is more than
      // 3, since 2 is just \r\n.

      if (left > 0) {
        
        // Copy remaining.
        memcpy(node->inbuffer, *last, left);

        // Reset pointers and PC
        node->in_buffer = left;
        next = NULL;
      } else {

        // Ignore the remainder. (It's just \r\n or a combination of those)
        //        node->inbuffer = 0;

      }

      // stop processing
      return 0;
    }


    // Check if we've hit the end.
    if ((*next == '\n') || (*next == '\r')) { /* Got a line */

      // yep, skip past all blank \r or \n.
      while((*next == '\n') || (*next == '\r')) 
        *(next++) = (char) 0;

      // and replace them with \0. (Terminate string).

      // If we've completely finished parsing the input
      // start again from the start. If not, "next" is ready for 
      // next time around.
#if 0
      printf("inbuf %d, next rel %d, next %s diff %d\n", node->inbuffer,
             (int)(next - node->buffer), next,
             (node->inbuffer - (int)(next - node->buffer) ));
#endif

      if (node->in_buffer - ((int)(next - (char *)node->inbuffer)) <= 0) {
        node->in_buffer = 0;
      }

      // Signal we have a valid string.
      return 1;
    }

    /* This next should never happen */
    /* But we check for it just incase. I've never seen it yet. */

    if ((next > (char *)node->inbuffer + DATA_BUFSZ)) {
      log_printf(LOG_ERROR, "Internal Error, most likely due to a bug\r\n");
      node->in_buffer = 0;
      return 0;
    }

    // Advance to next char
    next++;

  } while(1);

  // NOT REACHED
  return 0;  // avoid the warning
}




void site_command(site_t *node, char *line)
{

  log_sprintf(node, "<[%s]: %s\n", node->name, line);


  // Attempt to parse out the FTP protocol number here.

  if (isdigit(line[0]) && isdigit(line[1]) && isdigit(line[2]) &&
      line[3] == ' ') {

    log_printf(LOG_DEBUG, "*[%s] command %d\n", node->name, atoi(line));

    node->cmd_reply = atoi( line );
    node->cmd_queue[ node->cmd_offset++ ] = node->cmd_reply;
    if ( node->cmd_offset >= MAX_CMD_QUEUE ) node->cmd_offset = 0;

    
    if (node->cmd) {
      free(node->cmd);
      node->cmd = NULL;
    }

    node->cmd = mystrcpy(line);

  }


}





#if __STDC__ || WIN32
int site_printf(site_t *node, char const *fmt, ...)
#else
int site_printf(node, fmt, va_alist)
     site_t *node;
     char const *fmt;
     va_dcl
#endif
{
  va_list ap;
  char msg[1024];
  int result;

#if __STDC__ || WIN32
  va_start(ap, fmt);
#else
  va_start(ap);
#endif

  // Buffer overflow! No vsnprintf on many systems :(
  vsprintf(msg, fmt, ap);
  va_end(ap);

  
  result = site_output( node, msg, strlen( msg ));

  
  return result;
                 
}








int site_output( site_t *node, char *buffer, unsigned int len)
{
  int sent = 0; 

  // assert's

  buffer[len] = 0;
  log_sprintf(node, ">[%s]: %s\n", node->name, buffer);


  // First we check if there is anything already in the buffer, if
  // so, we attempt to send that (first?)

  if ( node->out_buffer ) {

    sent = sockets_write( node->socket, node->outbuffer, node->out_buffer);

    if (sent > 0) node->stats.bytes_to += sent;
    
    // did we manage to send it all?
    if (sent == node->out_buffer) {


      node->state.socket = SO_ACTIVE;
      node->out_buffer = 0;



    } else { // rats, we still have data to buffer.

      // Will the buffer be too full?  If so what choice do we have but to
      // either discard the info, or drop the connection with buffer overrun?

      if (node->out_buffer + (node->out_buffer - sent) > DATA_BUFSZ) {

        // drop the additional data :(
        log_printf(LOG_ERROR, "%08x buffer overflow -- data dropped\n", node->socket);

        node->out_buffer = 0;
        return 0;

      }


      memcpy( node->outbuffer, &node->outbuffer[ sent ], node->out_buffer - sent );

      node->out_buffer -= sent;
  
      node->state.socket = SO_BUFFERFULL;


    }

  }  // if node->out_buffer 


  // Check if we were passed new information to send
  if (!buffer || ( len == 0 ))
    return sent;



  // Attempt to send the new info, if we've already failed above
  // we don't attempt it again, just incase it would pass thru (and then
  // we'd get data out of order)

  sent = 0;


  if (node->state.socket != SO_BUFFERFULL)
    sent = sockets_write( node->socket, buffer, len);


  if (sent <= 0) {     // 0 is a connection closed, -1 is connection lost.

    // We can actually fairly safely ignore connection lost and closed here
    // since we will get a read() trigger due to it, and then close it.
    sent = 0;          // pretend we sent nothing, so we buffer it all.

  }

  node->stats.bytes_to += sent;


  if ((unsigned int)sent == len) {

    // We sent all that we wanted to send, so it's easy going.
    node->state.socket = SO_ACTIVE;

    return sent;

  }

  // Drat! We didn't send as much as we wanted to, buffer it.
  // Copy the remaining data to the output buffer, set the
  // counter and change our status.

  // Check if we would over flow.
  if (node->out_buffer + ( len - sent ) > DATA_BUFSZ) {

    // Drop the buffer :(
    log_printf(LOG_ERROR, "%08x Buffer overflow\n", node->socket);

    node->out_buffer = 0;

    return 0;

  }


  memcpy( &node->outbuffer[ node->out_buffer ], 
          &buffer[ sent ], len - sent );

  node->out_buffer += len - sent;
  
  node->state.socket = SO_BUFFERFULL;

  return sent;
}




int site_pasv(site_t *node)
{
  
  log_printf(LOG_DEBUG, "[%s] Fetching out PASV info from '%s'\n", node->name, node->cmd);


  if (!sockets_pasv(node->cmd, &node->pasv_host, &node->pasv_port)) {

    log_printf(LOG_ERROR, "[%s] Malformed 227 received from PASV :(\n", node->name);
    return 0;
  }


  log_printf(LOG_DEBUG, "[%s] PASV : %s:%d\n",  node->name, ul2a(node->pasv_host),
         node->pasv_port);


  return 1;

}



//
// 0 pending
// 1 success
// 2 failure
//
int site_connect(site_t *node)
{

  switch (node->state.pasv) {

  case SO_NONE:  // Fire off a connect
    node->pasv_socket = sockets_connect(node->pasv_host, node->pasv_port);

    if (node->pasv_socket == -1) {

      log_printf(LOG_ERROR, "[%s] Passive connection failed\n", node->name);
      return 2;

    }

    node->state.pasv = SO_PENDING;
    return 0;

    break;


  case SO_PENDING:
    io_set_read( node->pasv_socket );
    io_set_write( node->pasv_socket );
    return 0;


  case SO_ACTIVE:
    return 1;  // Success!!


  case SO_ERROR:
    if (node->pasv_socket > 0) {
      sockets_close(node->pasv_socket);
      node->pasv_socket = 0;
    }

    return 2;
    break;

  default:
    ;
  }

  return 0;

}


void site_disconnect(site_t *node)
{

  if ((node->socket > 0)) {

    sockets_close(node->socket);
    node->socket = 0;
    node->state.socket = SO_NONE;

  }

}




void site_waitclear(site_t *node)
{
  int i;

  for (i = 0 ; i < MAX_CMD_QUEUE; i++ ) {

    node->cmd_queue[i] = 0;

  }

}


int site_waitfor(site_t *node, unsigned int num)
{
  int i;

  for (i = 0 ; i < MAX_CMD_QUEUE; i++ ) {

    if (node->cmd_queue[i] == num) {

      //      node->cmd_queue[i] = 0;
      return 1;

    }

  }

  return 0;

}




int site_listen(site_t *node)
{

  switch (node->state.pasv) {

  case SO_LISTEN:
    io_set_read( node->pasv_socket );
    return 0;

  case SO_PENDING:
    io_set_read( node->pasv_socket );
    io_set_write( node->pasv_socket );
    return 0;


  case SO_ACTIVE:
    return 1;  // Success!!


  case SO_ERROR:
    if (node->pasv_socket > 0) {
      sockets_close(node->pasv_socket);
      node->pasv_socket = 0;
    }

    return 2;
    break;

  default:
    ;
  }

  return 0;
}
