// $Id: sockets.c,v 1.2 2000/03/16 14:03:54 lundman Exp $
// Lowest level socket manipulation routines.
// Jorgen Lundman October 10, 1999
// Copyright (c) 1999 HotGen Studios Ltd <www.hotgen.com>


#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <stdarg.h>
#include <sys/types.h>
#include <time.h>
#include <fcntl.h>
#include <signal.h>
#include <errno.h>
#if __STDC__
#include <stdarg.h>
#else
#include <varargs.h>
#endif

#ifndef WIN32 

#include <unistd.h>
#include <netinet/in.h>
//#include <netinet/in_systm.h>
//#include <netinet/ip.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <sys/socket.h>

#else

// Unfortunately winsock.h includes windows.h, so we try to get as little 
// of it as possible.
#define FD_SETSIZE 1024
#define WINDOWS_LEAN_AND_MEAN
#include <winsock.h>

#endif

// Some OS's don't define this, notibly SunOS 4 and below.
#ifndef INADDR_NONE
#define INADDR_NONE -1
#endif


#include "sockets.h"




// One-time initialisations required. (buffers, lists etc)
// These really just stubs incase of future need.
int sockets_init(void)
{



#ifdef WIN32
  // Only WIN32 needs to do something here
  // This is straight out of 
  WORD wVersionRequested;                               // WinSock specific
  WSADATA wsaData;					// WinSock specific
  int result; 
  
  wVersionRequested = MAKEWORD( 2, 2 ); 
  
  result = WSAStartup( wVersionRequested, &wsaData );
  
  if ( result ) {
    printf("Couldn't start WinSock :( \n");
    exit(1);
  }
  
  printf("WinSock version %d.%d Max %d\n", 
	 LOBYTE( wsaData.wVersion ), HIBYTE( wsaData.wVersion ), 
	 wsaData.iMaxSockets);
  
#endif


  return 0;
  
}


// Opposite to sockets_init.
void sockets_exit(void)
{


#ifdef WIN32
  WSACleanup( );
#endif


}



// Close an already open socket.
//
void sockets_close(int old_socket)
{

#ifdef WIN32
	closesocket(old_socket);
#else
	close(old_socket);
#endif

}





// (INCOMING)
// Open a new socket/port for incomming connections, set to nonblocking if 
// required.
//
// IN:  (int *) "port" number. If pre-set to 0 it will pick first available 
//      port and assign
// OUT: (int) new file-descriptor / SOCKET, or -1 for error. (Slight WIN32
//       override)
// OUT: (int *) "port" filled with local port number (host-order)
//
int sockets_make(int *port)
{
  int option_arg, new_socket;    // temporary argument int, and holder for 
                                 // new socket
  int result;                    // temporary holder for return code testing
  struct sockaddr_in local_sin;  // temporary socket Internet address holder


  // Request a new TCP socket in the Internet address family.	
  new_socket = socket( AF_INET, SOCK_STREAM, 0 );

  // Verify it was successful
#ifdef WIN32
  if (new_socket == INVALID_SOCKET) 
#else
  if (new_socket < 0 ) 
#endif
    {
      // Insert non-OS specific error handling here.
      perror("socket(): ");
      return -1;
    }

  

  // We set our new socket to be reusable, otherwise you end up with too 
  // many dead ports.
  option_arg = 1;
  // Unfortunately, most OS's use void * for the argument, except for
  // windows, where it is char *

  setsockopt(new_socket, SOL_SOCKET, SO_REUSEADDR, (void *)&option_arg, 
	     sizeof(option_arg));


  // If we want non-blocking mode, set it now
#ifdef NONBLOCKING_SOCKETS
  
#ifdef WIN32
  ioctlsocket(new_socket, FIONBIO, &option_arg);
#else
  fcntl(new_socket, F_SETFL, (fcntl(new_socket, F_GETFL) | O_NONBLOCK));
#endif

#endif


  // Setup the socket address structure
  local_sin.sin_family      = AF_INET;
  local_sin.sin_addr.s_addr = htonl(INADDR_ANY);
  local_sin.sin_port        = htons((unsigned short)*port); /* Any port */
  
  
  //	printf("bind()ing %s\n", ul2a(l_sin.sin_addr.s_addr));


  // Lets bind this new socket to a port
  result = bind( new_socket, (struct sockaddr *) &local_sin, 
		 sizeof(local_sin));

#ifdef WIN32

  if ( result == SOCKET_ERROR )
    {
      
      // Windows don't set the return code in "some" of their network calls.
      // This is a bit fudgy, but the remaining code stays portable
      errno = WSAGetLastError();
      
      perror("bind: ");
      closesocket(new_socket);
      
      return -1;
    }


#else
  
  if ( result )
    {
      
      perror("bind: ");
      close(new_socket);
      
      return -1;
    }

#endif
  
  
  // If the port was 0 and we picked a free one, now it's time to check
  // which port we were assigned.
  option_arg = sizeof(struct sockaddr_in);
  getsockname(new_socket, (struct sockaddr *) &local_sin, &option_arg);
  
  
  // Get the socket listening for new connections, the queue parameter is 
  // obsolete but is required for nostalgic reasons.
  listen(new_socket, 5);

  // Assigned it to the passed variable
  //*port = ntohs(local_sin.sin_port);
  
  // Finally, return the new and shining socket
  return new_socket;
}






// (INCOMING)
// accept a new connection on a socket.
//
// IN: socket where the new connection is on
// IN: <optional> (ULONG *) remote host's address (host-order)
// IN: <optional> (UINT  *) remote host's port    (host-order)
// OUT: the new socket, or -1 for error
//
int sockets_accept(int sock, unsigned long *remhost, int *remport)
{
  int new_socket;		// the new socket
  struct sockaddr_in from;	// temporary address holder
  int len = sizeof(from);	// temporary integer variable of address size


  // Call accept
  if (remhost)
    new_socket = accept(sock, (struct sockaddr *)&from, &len);
  else
    new_socket = accept(sock, NULL, NULL);
  
  
#ifdef WIN32
  if (new_socket == INVALID_SOCKET) return -1;
#else
  if (new_socket < 0) return -1;
#endif
  
  
  // bcopy is non-standard so we change to ansi memcpy
  // bcopy(&from.sin_addr.s_addr, remhost, sizeof(unsigned long));
  
  // copy over the actual IP, this *will* break with IPng.
  if (remhost)
    {
      memcpy((void *)remhost, (void *)(&from.sin_addr.s_addr), 
	     sizeof(unsigned long));
      *remhost = ntohl(*remhost);
      
      // and port, in host-order
      *remport = ntohs(from.sin_port);
    }
  
  
#if !defined ( LINUX ) && defined ( DATA_BUFSZ )
  
  /*
   *
   * This is by no means crucial for the ftpd to work, but will make it run
   * smoother. It is part of POSIX but lame OS, like Linux, decide only to
   * implement POSIX half-heartedly. But as always, the popular OS' are never
   * the good OS'. (See ms-dos, Win95, Linux etc)
   *
   */
  
  
  // Set low-water mark, which is still not implemented by Linux. 
  len = DATA_BUFSZ;
  setsockopt(new_socket, SOL_SOCKET, SO_SNDLOWAT, &len, sizeof(len));
  
#endif
  
  
  // Set non blocking
#ifdef WIN32
  // re-using local variables
  len = 1;
  ioctlsocket(new_socket, FIONBIO, &len);
#else
  fcntl(new_socket,F_SETFL,(fcntl(new_socket,F_GETFL)|O_NONBLOCK));
#endif
  
  
  // return new socket just connected
  return new_socket;
}





// (OUTGOING)
// Establish a new connection to a remote host
//
// IN: (ULONG) address of remote host
// IN: (UINT)  port of remote host
// OUT: the new socket, or -1 for error
//
int sockets_connect(unsigned long remote_host, unsigned int remote_port)
{
  struct sockaddr_in local_sin;  // temporary address holder
  int new_socket;                // the new socket to be returned
  int result;
#if !defined (LINUX)
  int option_arg;
#endif
  
  // Allocate new socket
  new_socket = socket( AF_INET, SOCK_STREAM, 0 );
  
  
#ifdef WIN32
  if (new_socket == INVALID_SOCKET)
#else
    if (new_socket < 0) 
#endif
      {
	perror("socket_connect(socket())");
	return -1;
      }
  
  
  
  // Fill in the members of the address structure and byte swap
  local_sin.sin_port = htons((unsigned short)remote_port);
  //local_sin.sin_port = (unsigned short)remote_port;
  local_sin.sin_family = AF_INET;
  
  // Byte swap host
  remote_host = htonl(remote_host);
  
  // copy over the address
  memcpy(&local_sin.sin_addr, &remote_host, sizeof(remote_host)) ;
  
  
#if !defined ( LINUX ) && defined (DATA_BUFSZ)
  // Set low water mart on all but Linux
  option_arg = DATA_BUFSZ;
  setsockopt(new_socket, SOL_SOCKET, SO_SNDLOWAT, &option_arg, 
	     sizeof(option_arg));
  
#endif
  
  
  // Enable non-blocking mode if required
#ifdef NONBLOCKING_SOCKETS
  
#ifdef WIN32
  option_arg = 1;
  ioctlsocket(new_socket, FIONBIO, &option_arg);
#else
  fcntl(new_socket,F_SETFL,(fcntl(new_socket,F_GETFL)|O_NONBLOCK));
#endif
  
#endif // NONBLOCKING_SOCKETS
  
  
  // Attempt to connect to remote host
  result = connect(new_socket,(struct sockaddr *)&local_sin, 
		   sizeof(local_sin));
  
  
  // Check the error, some OS's return an error to say it is in progress, 
  // others
  // return just success, so we have to check for both eventualities.
#ifdef WIN32
  errno = WSAGetLastError();
  
  if( (result == SOCKET_ERROR) && ( errno != WSAEINPROGRESS)  && (errno != WSAEWOULDBLOCK))
    {
      /* Doh, failed. */
      closesocket(new_socket);
      return -1;
    }
  
#else
  
  if( result && ( errno != EINPROGRESS ) && ( errno != EWOULDBLOCK)) 
    {
      /* Doh, failed. */
      close(new_socket);
      return -1;
    }
  
#endif
  
  //
#if 0
  printf("Experimental settings\n");
  {
    int on = 1;
    if (setsockopt(new_socket, SOL_SOCKET, SO_OOBINLINE, (char *)&on, sizeof(on))
	< 0) {
      perror("setsockopt");
    }
  }
  {
    int tos = IPTOS_LOWDELAY;
    if (setsockopt(new_socket, IPPROTO_IP, IP_TOS, (char *)&tos,
		   sizeof(int)) < 0)
      perror("setsockopt TOS (ignored)");
  }
#endif  



  // Return the new socket. If we are in nonblocking, this is NOT connected 
  // yet and may
  // still fail
  return new_socket;
  
}




// Attempt to retrieve more data from a socket
//
// IN: socket, buffer area, max read (a-la read(). )
// OUT: number of bytes written to buffer, 0 closed gracefully, -1 lost.
//
int sockets_read(int fd, char *buffer, unsigned int len)
{
  
  // We use read() for Unix platforms as it's the most portable
  // even tho most systems also have recv().  Win32 can't call
  // read() and amusing things happen when you try.
  
#ifdef WIN32
  return recv(fd, buffer, len, 0);
#else
  return read(fd, buffer, len);
#endif
}




// Attempt to transmit more data to a socket
//
// IN: socket, buffer area, number of bytes to send.
// OUT: number of bytes sent, 0 closed gracefully, -1 lost.
//
int sockets_write(int fd, char *buffer, unsigned int len)
{
#ifdef WIN32
  return send(fd, buffer, len, 0);
#else
  return write(fd, buffer, len);
#endif
}




// 
// Standard formattable printing routine. Testing only, no buffer checks!
//
// IN:  socket to transmit to
// IN:  format and arguments a'la printf
// OUT: bytes written or -1 for error, or 0 for connection gracefully closed.
//
#if __STDC__
int sockets_printf(int fd, char const *fmt, ...)
#else
int sockets_printf(fd, fmt, va_alist)
     int fd;
     char *fmt;
     va_dcl
#endif
{
  va_list ap;
  char msg[1024];
	int result;

#if __STDC__
  va_start(ap, fmt);
#else
  va_start(ap);
#endif

  // Buffer overflow! No vsnprintf on many systems :(
  vsprintf(msg, fmt, ap);
  va_end(ap);

#ifdef WIN32

  result = send(fd, msg, strlen(msg), 0);
  
#else
  
  result = write(fd, msg, strlen(msg));
  
#endif
  
  return result;
  
}


unsigned long sockets_gethost(char *host)
{
  unsigned long result;
  struct hostent *h;
    

  // Is string in dot notation?
  result = inet_addr( host );

  if (result == -1) {
 
    // Nope, try to look it up as a host name
    h = gethostbyname(host);

    if (!h) {
      // Failed to look up hostname
      printf("Failed to lookup hostname %s\n", host);
      return 0;  // 0 is a valid IP :(
    }
    
    memcpy(&result, h->h_addr, sizeof(result));

  }

  result = ntohl(result);

  return result;

}




char *ul2a(unsigned long q)
{
  static char host[64];

  q = ntohl(q);  // Bring it to host order

  sprintf(host, "%u.%u.%u.%u",
          ((unsigned char *)&q)[0], /* Assume network order */
          ((unsigned char *)&q)[1],
          ((unsigned char *)&q)[2],
          ((unsigned char *)&q)[3]);

  return host;
}





int sockets_pasv(char *line, unsigned long *host, unsigned short int *port)
{
  unsigned char *pp;
  int IPA[4], PORTA[2];

  pp = strchr(line, '(');

  if (!pp) {
    return 0;
  }

  if (sscanf(&pp[1], "%d,%d,%d,%d,%d,%d", &IPA[0], &IPA[1], &IPA[2], &IPA[3],
             &PORTA[0], &PORTA[1]) != 6) {
    return 0;
  }

  pp = (unsigned char *)(host);
  pp[0] = (unsigned char) IPA[0];
  pp[1] = (unsigned char) IPA[1];
  pp[2] = (unsigned char) IPA[2];
  pp[3] = (unsigned char) IPA[3];
  pp = (unsigned char *)(port);
  pp[0] = (unsigned char) PORTA[0];
  pp[1] = (unsigned char) PORTA[1];

  *host = ntohl(*host);
  *port = ntohs(*port);

  return 1;
}


unsigned long sockets_getsockname(int fd)
{
  struct sockaddr_in par;
  int len;
  unsigned long result;

  len = sizeof(par);

  if (getsockname(fd, (struct sockaddr *)&par, &len) < 0) {
      
    //bcopy(&localhost, &(t->port_host), sizeof(localhost));
    return INADDR_NONE;

    } 

  memcpy(&result, &par.sin_addr, sizeof(result));
    
  return result;

}

