#include  <stdio.h>
#include <stdlib.h>

#include "mirror.h"
#include "state.h"
#include "sockets.h"
#include "io.h"
#include "site.h"
#include "collect.h"
#include "ctype.h"
#include "misc.h"
#include "compare.h"
#include "fxp.h"
#include "pasv.h"
#include "cps.h"
#include "log.h"
#include "date.h"
#include "stats.h"




static int state_skip_size = 0;





void state_reset(site_t *node)
{

  node->state.state = ST_NONE;
  node->state.substate = 0;

  node->idle = io_global_time;

  pasv_freedir(&(node->mirror_paths));
  
  node->mirror_current = NULL;

}


void state_process(mirror_t *mirror, site_t *node)
{

  switch( node->state.state ) {


  case ST_NONE:
    state_none( node );
    break;

  case ST_LOGIN:
    state_login( mirror, node );
    break;

  case ST_COLLECT:
    state_collect( mirror, node );
    break;

  case ST_COMPARE:
    state_compare( mirror, node );
    break;

  case ST_MIRROR:
    state_mirror( mirror, node );
    break;

  case ST_DONE:
    state_done( mirror, node );
    break;

  case ST_RETRY:
  case ST_DISCONNECT:
    state_retry( mirror, node);
    break;

  }



  if (node->state.socket == SO_ACTIVE) {

    io_set_read (node->socket);

  }
  else if (node->state.socket == SO_BUFFERFULL)
    io_set_write(node->socket);
  else return;
  
  if (node->state.pasv == SO_ACTIVE) {

    io_set_read(node->pasv_socket);

  }

  // Check idle.

  // Use different timeout for DATA if we are FXPing at the time
  if ((mirror->A_host.state.state == ST_MIRROR) &&
      (mirror->A_host.state.substate == 18)) {

    if ((io_global_time - node->idle) > node->timeout_transfer) { 


      log_printf(LOG_ERROR, "Timeout detected during FXP transfer\n");
      state_set(node, ST_RETRY);

    }



  } else if (node->state.state != ST_RETRY) {

    if ((io_global_time - node->idle) > node->timeout) {

      log_printf(LOG_ERROR, "Timeout detected %d %d\n", node->state.state, 
	     node->state.substate);


      

      state_set(node, ST_RETRY);

    }

  }

}




void state_set(site_t *node, enum state_type state)
{
  
  // We are changing to new state
  switch(state) {

  case ST_RETRY:
  case ST_DISCONNECT:
    node->state.substate = 0;
    state = ST_RETRY;
    break;

  case ST_LOGIN:
    node->state.substate = 0;
    break;

  case ST_COLLECT:
    node->state.substate = 0;
    break;

  case ST_COMPARE:
    node->state.substate = 0;
    break;

  case ST_MIRROR:
    node->state.substate = 0;
    break;

  case ST_DONE:
    node->state.substate = 0;
    break;


  default:
    ;  // MSVC has a bug, it'll give error with a ; here
  }


  node->state.state = state;


}















void state_none(site_t *node)
{
  // Check socket status

  switch (node->state.socket ) {

  case SO_NONE:
    // Connect to remote host.
    log_printf(LOG_DEBUG, "Connecting to %s\n", node->name);


    node->socket = sockets_connect( node->ip, node->port );


    if (node->socket == -1) {

      log_printf(LOG_DEBUG, "Failed to connect to %s -- retry\n", node->name);

      state_set(node, ST_RETRY);

      return;

    }

    node->state.socket = SO_PENDING;

    break;


  case SO_PENDING:
    
    if (node->socket == -1) {    // shouldn't happen
      
      state_set(node, ST_RETRY);
      
      return;
    }
    
    io_set_read (node->socket);
    io_set_write(node->socket);
    
    break;


    
  case SO_ACTIVE:
    state_set(node, ST_LOGIN);
    break;
    

    
  case SO_BUFFERFULL:
    io_set_write(node->socket);
    break;
    
  default:
    ;
  }


}





void state_login(mirror_t *mirror, site_t *node)
{



  //  printf("FTP num seen %d\n", node->cmd_reply);



  switch(node->state.substate) {

  case 0:
    // Wait until we've seen 220
    if (site_waitfor(node,  220))
      node->state.substate++;
    break;



  case 1:
    // Send USER
    site_printf(node, "USER %s\r\n",
		node->user);
    site_printf(node, "PASS %s\r\n",
		node->pass);
    node->state.substate++;
    break;

  case 2:
    // Here we wait for authentication to fail or succeed

    if (site_waitfor(node,  530)) {  // Failure

      log_printf(LOG_ERROR, "[%s] Failed to login\n", node->name);

      state_set(node, ST_RETRY);
      break;
      

    } else if (site_waitfor(node,  230)) { // success

      log_printf(LOG_DEBUG, "[%s] Login successful...\n", node->name);

      site_printf(node, "TYPE I\r\n");

      node->state.substate++;
      
      break;

    }
    break;

  case 3:

#ifndef ONE_SITE
    if ((mirror->A_host.state.substate == 3) &&
	(mirror->B_host.state.substate == 3)) {
#else
      if (1) {
#endif

      collect_setpaths( mirror );

      break;

    }

    break;

  }

}






void state_collect(mirror_t *mirror, site_t *node)
{
  int ret;
  int pickone;
  int new_socket;
  //  printf("FTP num seen %d  %d\n", node->cmd_reply, node->state.substate);


  switch(node->state.substate) {

  case 0:
    site_waitclear(node);

    // Change directory
    site_printf(node, "CWD %s\r\n", node->current_cwd);
    node->state.substate++;
    break;


  case 1:
    // Wait for success or failure of CWD
    if (site_waitfor(node,  250)) {
      log_printf(LOG_DEBUG, "[%s] CWD OK.\n", node->name);
      node->state.substate++;
      break;
    }

    if (site_waitfor(node,  550)) {
      log_printf(LOG_ERROR, "[%s] CWD failed -- stop\n", node->name);
      state_set(node, ST_RETRY);
      break;
    }
    break;

  case 2:
    // Start a PORT or PASV session.
    if (node->passive == YN_NO) {

      node->state.substate = 10;    // Go to state 10 for port

    } else {

      node->state.substate = 20;    // Go to state 20 for passive

    }
    break;



    //
    // PORT
    //

  case 10:
    //    log_printf(LOG_ERROR, "PORT method not yet implemented\n");
    //    state_set(node, ST_RETRY);
    node->state.substate++;
    break;

  case 11:
    // fetch a new port from sockets layer
    pickone = 0;

    if ((new_socket = sockets_make(&pickone)) == -1) {

      log_printf(LOG_ERROR, "Couldn't open a local data socket\n");
      state_set(node, ST_RETRY);
      break;
    }

    log_printf(LOG_DEBUG, "New port is %d\n", pickone);


    node->pasv_socket = new_socket;


    //    node->pasv_port = ntohs((unsigned short int)pickone);
    node->pasv_port = (unsigned short int)pickone;

    //node->pasv_port = (unsigned short int)pickone;
    node->pasv_host = sockets_getsockname( node->socket );

    site_waitclear(node);

    fxp_port(node);

    node->state.substate++;
    break;

  case 12:
    // WAIT for 200 PORT
    if (site_waitfor(node, 200))
      node->state.substate++;
    break;

  case 13:
    // Issue list
    node->state.pasv = SO_LISTEN;

    site_printf(node, "LIST -l\r\n");
    node->state.substate++;
    break;


  case 14:
    // Wait for data connection
    switch( site_listen( node )) {

    case 1:  // Connected
      node->state.substate++;
      break;

    case 2:  // failed
      log_printf(LOG_ERROR, "[%s] data connection failed\n", node->name);
      state_set(node, ST_RETRY);
      break;

    }
    break;

  case 15:
    log_printf(LOG_DEBUG, "[%s] data connected\n", node->name);

    node->state.substate = 24;  // wait for 150, then 226 then...
    break;






    //
    // PASV
    //

  case 20:
    // Send passive string.
    site_printf(node, "PASV\r\n");
    node->state.substate++;
    break;


  case 21:
    // Wait for reply to PASV

    if (site_waitfor(node,  227)) {

      if (!site_pasv(node)) {
	state_set(node, ST_RETRY);
	break;
      }

      // Passive successful, connect first.

      
      node->state.substate++;
      break;

    }

    if (site_waitfor(node,  425)) {
      state_set(node, ST_RETRY);
      break;
    }

    break;


  case 22:

    ret = site_connect(node);

    // 0=inprogress, 1=success, 2=failure

    if (ret == 1) {  // connected

      log_printf(LOG_DEBUG, "[%s] Connected to passive port\n", node->name);
      node->state.substate++;
      break;

    }


    if (ret == 2) {  // failed

      log_printf(LOG_ERROR, "[%s] Failed to connect to passive\n", node->name);
      state_set(node, ST_RETRY);
      break;

    }

    break;



  case 23:
    //  Connected, send list command.
    
    site_printf(node, "LIST -l\r\n");
    node->state.substate++;
    break;


  case 24:
    // send ls, wait for 150 start
    if (site_waitfor(node, 150)) { // We've started
      //      site_printf(node, "SYST\r\n");
      node->state.substate++;
    }
    break;


  case 25:
    // wait for 226 finish line
    if (site_waitfor(node, 226)) { // We've started
      node->state.substate++;
    }
    break;

  case 26:
    // Wait for data session to close.

    if (node->state.pasv == SO_NONE) {

      sockets_close(node->pasv_socket);
      node->pasv_socket = 0;

      // We are done collecting
      state_set(node, ST_COMPARE);

    }

    if (node->state.pasv == SO_ERROR) {

      sockets_close(node->pasv_socket);
      node->pasv_socket = 0;

      // Failed to get directory
      state_set(node, ST_RETRY);

    }
    break;

  }

}





void state_compare(mirror_t *mirror, site_t *node)
{
  // here for the first time we wait for both hosts to 
  // reach this mode.
  dir_t *runner;


  if ((mirror->A_host.state.state == ST_COMPARE) && 
      (mirror->B_host.state.state == ST_COMPARE)) {

    log_printf(LOG_DEBUG, "Both reached ST_COMPARE\n");

#if 0
    log_printf(LOG_DEBUG, "These are the files/directories that we found on each host\n");

    log_printf(LOG_DEBUG, "[%s]\n", mirror->A_host.name);

    for (runner = mirror->A_host.current_dir; runner; runner = runner->next) {
      log_printf(LOG_DEBUG, "%3d | %8lu | %30s | %s", runner->type, runner->size, 
	     runner->name, ctime(&runner->date));

    }

    log_printf(LOG_DEBUG, "[%s]\n", mirror->B_host.name);

    for (runner = mirror->B_host.current_dir; runner; runner = runner->next) {
      log_printf(LOG_DEBUG, "%3d | %8lu | %30s | %s", runner->type, runner->size, 
	     runner->name, ctime(&runner->date));

    }

    log_printf(LOG_DEBUG, "done.\n");
#endif



    // If skip_initial is set, we just fetch out the newest dir
    // date_stamp it, and exit.
    if (mirror->A_host.skip_initial) {

      // Find newest
      for (runner = mirror->A_host.current_dir; 
	   runner; 
	   runner = runner->next) {

	if (!runner->next) {
	  date_stamp(mirror, &mirror->A_host, runner);

	  

	  if (mirror->A_host.current_path->next) {

	    collect_setpaths(mirror);
	    return;

	  }


	  log_printf(LOG_DEBUG, "Date stamped -- exit\n");

	}
      }

      state_set(&mirror->A_host, ST_DONE);
      state_set(&mirror->B_host, ST_DONE);

      return;
    }






    // Build nodes of which things we should xfer
    compare_dirs(mirror, &(mirror->A_host.mirror_paths));


    state_set(&mirror->A_host, ST_MIRROR);
    state_set(&mirror->B_host, ST_MIRROR);


    // !FIXME
    // Free all nodes in current_dir

    
  }

}







//
// Take off the first node in ->mirror_paths
// process it.
// file - xfer then remove
// dir  - enter and list, creating new entries to mirror_paths
//        recurse - eventually remove
//
void state_mirror(mirror_t *mirror, site_t *node)
{
  int ret;
  int pickone;
  int new_socket;
  time_t diff;

  if (!mirror->A_host.mirror_paths) {

    log_printf(LOG_DEBUG, "[%s] Transfer queue exhausted - time for a break.\n",
	       mirror->A_host.name);
    
    log_printf(LOG_DEBUG, "Advancing to next mirror path\n");
    
    pasv_freedir(&(mirror->A_host.mirror_paths));
    pasv_freedir(&(mirror->B_host.mirror_paths));
    
    mirror->A_host.mirror_current = NULL;
    mirror->B_host.mirror_current = NULL;
    
    mirror->A_host.retries_done = 0;
    mirror->B_host.retries_done = 0;
    
    collect_setpaths( mirror );

    return;
  }
    

  // In here we only deal with A_host
  // unless it's substate between 60-100 for directory
  // fetching
  if (node != &(mirror->A_host) &&
      (mirror->B_host.state.substate < 60)) return;


#if 0
  log_printf(LOG_ ,"mirror %d %s\n", mirror->A_host.state.substate,
	 mirror->A_host.mirror_paths->name);
#endif


  // process
  switch (node->state.substate) {

  case 0:
    site_waitclear(node);
    state_advance(mirror);
 

    // If we've changed state in advance, leave
    if (mirror->A_host.state.state != ST_MIRROR) return;


    if (!mirror->A_host.mirror_current)
      state_advance(mirror);


    mirror->A_host.state.substate++;
    break;


  case 1:
    // process node
    if (mirror->A_host.mirror_current->type == DT_FILE) {  
      mirror->A_host.state.substate = 10;  // Deal with file
      break;
    }

    mirror->A_host.state.substate = 40;  // Deal with dir

    break;






    //
    // File
    //




  case 10:
    log_printf(LOG_DEBUG, "Transfering file %s\n", mirror->A_host.mirror_current->name);

#if 0
    site_printf(&mirror->A_host, "PWD\r\n");
    site_printf(&mirror->B_host, "PWD\r\n");
#endif

    site_waitclear(node);


    if (state_skip_size) {
      // Skip the SIZE command if we already know it doesn't exist
      mirror->B_host.restart = 0;
      mirror->A_host.state.substate = 12;
      break;

    }

    // Check if the file exists by asking for its SIZE
    site_printf(&mirror->B_host, "SIZE %s\r\n",
		mirror->A_host.mirror_current->name);
    mirror->A_host.state.substate++;
    break;


  case 11:
    // When talking to a Linux FTPD I get this, how scary is that.
    // -rw-r--r--   1 ftp      ftp             5 Nov 10 11:42 Return
    // ftp> quote SIZE Return
    // 213 6

    // Only on Linux ... sigh
    
    // Seems SunFTPD doesn't have the command SIZE


    if (mirror->B_host.cmd_reply == 550) {
      // No such file, start from scratch.

      mirror->B_host.restart = 0;
      mirror->A_host.state.substate++;
      break;

    }


    if (mirror->B_host.cmd_reply == 500) {
      // Unknown command, we'll assume to start from scratch
      log_printf(LOG_DEBUG, "Server doesn't know SIZE command, resuming not possible\n");

      state_skip_size = 1;

      mirror->B_host.restart = 0;
      mirror->A_host.state.substate++;
      break;

    }


    if (mirror->B_host.cmd_reply == 213) {
      // Fetch out resume value.

      if (sscanf(mirror->B_host.cmd, "213 %lu", 
		 &(mirror->B_host.restart)) != 1)
	mirror->B_host.restart = 0;

      mirror->A_host.state.substate++;

      
      if (mirror->B_host.restart >= mirror->A_host.mirror_current->size) {
	log_printf(LOG_DEBUG, "[%s] Target %s (%lu -> %lu) is same or bigger -- skipping\n",
	       mirror->B_host.name,
	       mirror->A_host.mirror_current->name,
	       mirror->A_host.mirror_current->size,
	       mirror->B_host.restart);
	
	mirror->A_host.state.substate = 0;
	break;
      }
      

      log_printf(LOG_DEBUG, "Resuming file %s from %lu\n", 
	     mirror->A_host.mirror_current->name, mirror->B_host.restart);

    }
    break;


  case 12:
    
    fxp_init(mirror);
    
    mirror->A_host.state.substate++;
    break;
    

  case 13:
    // Waiting for both site to agree on pasv/port
    switch (fxp_initdone(mirror)) {

    case 1:
      mirror->A_host.state.substate++;
      break;

    case -1:
      state_set(&mirror->A_host, ST_RETRY);
      state_set(&mirror->B_host, ST_RETRY);

      break;
    default:
      ;
    }
    break;




  case 14:
    if (!mirror->B_host.restart) {   // If we are to resume, send REST packet
      // first
      
      mirror->A_host.state.substate = 16;
      break;
    }

    site_printf(&mirror->A_host, "REST %lu\r\n",
		mirror->B_host.restart);

    //    log_printf(LOG_DEBUG, "Warning - Programmer could not find FTP without resume to test failure code.\n");

    mirror->A_host.state.substate++;
    break;


  case 15:
    // wait for successful rest command.

    if (mirror->A_host.cmd_reply == 350) {
      // we are resuming, yay!

      mirror->A_host.state.substate++;
      break;
    }

    if ((mirror->A_host.cmd_reply == 500) ||
	(mirror->A_host.cmd_reply == 502)) {

      log_printf(LOG_DEBUG, "[%s] REST command refused. Resuming not possible\n",
	     mirror->A_host.name);

      mirror->B_host.restart = 0;
      mirror->A_host.state.substate++;
      break;

    }
    break;


  case 16:
    // port/pasv done. Send retr/stor pairs
    site_waitclear(&mirror->A_host);
    site_waitclear(&mirror->B_host);

    fxp_xfer(mirror);
    mirror->A_host.state.substate++;
    break;


  case 17:
    // Wait for both xfer's to commence.

    // check for failure

    if (site_waitfor(&mirror->A_host, 550) /* ||
	site_waitfor(&mirror->B_host, 550) */ ) {
      
      //      mirror->A_host.state.substate++;
      state_set(&mirror->A_host, ST_RETRY);
      state_set(&mirror->B_host, ST_RETRY);

    }


    // Some sites give this error sometimes, we need to just retry it.
    if (site_waitfor(&mirror->A_host, 426)) {
      state_set(&mirror->A_host, ST_RETRY);
      state_set(&mirror->B_host, ST_RETRY);
    }


    if (site_waitfor(&mirror->A_host, 553) ||
	site_waitfor(&mirror->B_host, 553) ||
	site_waitfor(&mirror->B_host, 550) ||
	site_waitfor(&mirror->A_host, 425) ||
	site_waitfor(&mirror->B_host, 425)) {

      // Clear REST incase it was sent
      if (mirror->B_host.restart) { 
	
	mirror->B_host.restart = 0;
	site_printf(&mirror->A_host, "REST 0\r\n");
	
      }


      mirror->A_host.state.substate = 999;
      break;

    }
    
    if (site_waitfor(&mirror->A_host, 150) &&
	site_waitfor(&mirror->B_host, 150)) {

      log_printf(LOG_DEBUG, "Transfer started...\n");

      log_printf(LOG_DEBUG, "Avg KB/sec  |Last KB/sec  |File ETA    |Dir ETA     |Est. %%\n");

      node->start = io_global_time;

      mirror->A_host.state.substate++;
    }
    break;

  case 18:
    // we are xfering.
    // wait for 226
    if (site_waitfor(&mirror->A_host, 226)) {

      mirror->A_host.state.substate++;
      mirror->A_host.idle = io_global_time;
      mirror->B_host.idle = io_global_time;

      log_printf(LOG_DEBUG, "\n");

    }
    break;


  case 19:
    if (site_waitfor(&mirror->B_host, 226))
      mirror->A_host.state.substate++;
    break;
    

  case 20: 
    // done
    diff = io_global_time - node->start;
    if (!diff) diff = 1;


    node->stats.bytes_site_to_site += 
      (unsigned MY_LONG64)mirror->A_host.mirror_current->size ;

    node->stats.total_files++;


    log_printf(LOG_DEBUG, "Done with file. %u seconds at %2.2fKB/sec\n",
	   (unsigned int) diff,
	   ((float)mirror->A_host.mirror_current->size) / ((float)diff)
	   / 1024.0);


    log_printf(LOG_MIRROR, "%s (%u sec at %7.2fKB/sec)\n",
	       mirror->A_host.mirror_current->name,
	       (unsigned int) diff,
	       ((float)mirror->A_host.mirror_current->size) / ((float)diff)
	       / 1024.0);


    // Only do a cps reading if it actually took some time.
    if (diff > 5)
      cps_new(((float)mirror->A_host.mirror_current->size) / ((float)diff));


    //pasv_removefirst(&(mirror->A_host.mirror_current));
    mirror->A_host.state.substate = 0;
    break;







    //
    // Directory
    //




  case 40:
    log_printf(LOG_DEBUG, "Processing dir %s\n", mirror->A_host.mirror_current->name);

    log_printf(LOG_MIRROR, "%s (Starting on directory)\n", mirror->A_host.mirror_current->name);


    site_waitclear(node);

    // Send CWD to site A
    site_printf(&mirror->A_host, "CWD %s\r\n", 
		mirror->A_host.mirror_current->name);

    mirror->A_host.state.substate++;
    break;



  case 41:
    // Wait for success of CD.
    if (site_waitfor(&mirror->A_host, 250)) {
      log_printf(LOG_DEBUG, "[%s] CWD OK\n", mirror->A_host.name);
      mirror->A_host.state.substate++;
      break;
    }

    if (site_waitfor(&mirror->A_host,550)) {
      mirror->A_host.state.substate = 999;   // error
      log_printf(LOG_ERROR, "[%s] CWD failed -- skipping\n", mirror->A_host.name);
      break;
    }
    break;




  case 42:
    // CD on host_B, if it fails, create dir.
    site_waitclear(&mirror->B_host);


    if (mirror->A_host.mirror_current->dest_name)
      site_printf(&mirror->B_host, "CWD %s\r\n", 
		  mirror->A_host.mirror_current->dest_name);
    else
      site_printf(&mirror->B_host, "CWD %s\r\n", 
		  mirror->A_host.mirror_current->name);

    mirror->A_host.state.substate++;
    break;


  case 43:
    // Wait for success of CD.
    if (site_waitfor(&mirror->B_host,250)) {
      log_printf(LOG_DEBUG, "[%s] CWD OK\n", mirror->B_host.name);
      mirror->A_host.state.substate = 48;
      break;
    }

    if (site_waitfor(&mirror->B_host, 550)) {

      // If the CD was the dest_name, go back and try it just name
      // and eventually MKDIR as that.
      if (mirror->A_host.mirror_current->dest_name) {

	free(mirror->A_host.mirror_current->dest_name);
	mirror->A_host.mirror_current->dest_name = NULL;

	mirror->A_host.state.substate = 42;
	break;

      }

      mirror->A_host.state.substate++;
      break;
    }
    break;


  case 44:
    site_waitclear(&mirror->B_host);

    // Attempt to create the dir
    site_printf(&mirror->B_host, "MKD %s\r\n", 
		mirror->A_host.mirror_current->name);
    mirror->A_host.state.substate++;
    break;

  case 45:
    // Wait for MKDIR success or failure.
    // We will also accept directory already exists

    if (site_waitfor(&mirror->B_host,257) ||
	site_waitfor(&mirror->B_host,521) ||   // Sent by linux
	site_waitfor(&mirror->B_host,550)) {   // Sent by others

      log_printf(LOG_DEBUG, "[%s] MKDIR\n", mirror->B_host.name);
      mirror->A_host.state.substate++;
      break;
    }

  case 46:
    site_waitclear(&mirror->B_host);
    // CD on host_B, if it fails, create dir.
    site_printf(&mirror->B_host, "CWD %s\r\n", 
		mirror->A_host.mirror_current->name);

    mirror->A_host.state.substate++;
    break;


  case 47:
    // Wait for success of CD.
    if (site_waitfor(&mirror->B_host,250)) {
      log_printf(LOG_DEBUG, "[%s] CWD OK\n", mirror->B_host.name);
      mirror->A_host.state.substate++;
      break;
    }

    if (site_waitfor(&mirror->B_host,550)) {
      log_printf(LOG_ERROR, "[%s] Failed to MKDIR/CWD\n", mirror->B_host.name);

      site_printf(&mirror->A_host, "CWD ..\r\n");

      mirror->A_host.state.substate = 999;   // error
      break;
    }
    break;


  case 48:
    // Both hosts are in subdir, finally.
    log_printf(LOG_DEBUG, "Done\n");
    //    site_printf(&mirror->B_host, "MKD .lmirror-0.01\r\n");
    //    site_printf(&mirror->B_host, "APPE .lmirror-0.01\r\n");
    node->stats.total_directories++;


  case 50:
    // Start a PORT or PASV session.
    if (mirror->A_host.passive == YN_NO) {

      mirror->A_host.state.substate = 60;

    } else {

      mirror->A_host.state.substate = 80;

    }

    if (mirror->B_host.passive == YN_NO) {

      mirror->B_host.state.substate = 60;

    } else {

      mirror->B_host.state.substate = 80;

    }
    break;



    //
    // PORT
    //

  case 60:
    //    log_printf(LOG_ERROR, "PORT method not yet implemented\n");
    //    node->state.substate = 999;    // error
    node->state.substate++;
    break;


  case 61:
    // fetch a new port from sockets layer
    pickone = 0;

    if ((new_socket = sockets_make(&pickone)) == -1) {

      log_printf(LOG_ERROR, "Couldn't open a local data socket\n");
      state_set(node, ST_RETRY);
      break;
    }

    log_printf(LOG_DEBUG, "New port is %d\n", pickone);


    node->pasv_socket = new_socket;


    //node->pasv_port = ntohs((unsigned short int)pickone);
    node->pasv_port = (unsigned short int)pickone;
    node->pasv_host = sockets_getsockname( node->socket );

    site_waitclear(node);

    fxp_port(node);

    node->state.substate++;
    break;

  case 62:
    // WAIT for 200 PORT
    if (site_waitfor(node, 200))
      node->state.substate++;
    break;

  case 63:
    // Issue list
    node->state.pasv = SO_LISTEN;

    site_printf(node, "LIST -l\r\n");
    node->state.substate++;
    break;


  case 64:
    // Wait for data connection
    switch( site_listen( node )) {

    case 1:  // Connected
      node->state.substate++;
      break;

    case 2:  // failed
      log_printf(LOG_ERROR, "[%s] data connection failed\n", node->name);
      state_set(node, ST_RETRY);
      break;

    }
    break;

  case 65:
    log_printf(LOG_DEBUG, "[%s] data connected\n", node->name);

    node->state.substate = 84;  // wait for 150, then 226 then...
    break;




    //
    // PASV
    //

  case 80:
    site_waitclear(node);

    // Send passive string.
    site_printf(node, "PASV\r\n");
    node->state.substate++;
    break;


  case 81:
    // Wait for reply to PASV

    if (site_waitfor(node,  227)) {

      if (!site_pasv(node)) {
	state_set(node, ST_RETRY);
	break;
      }

      node->state.substate++;
      break;

    }

    if (site_waitfor(node,  425)) {
      node->state.substate = 999;    // error
      break;
    }
    
    break;


  case 82:
    // Passive successful, connect first.

    ret = site_connect(node);

    // 0=inprogress, 1=success, 2=failure

    if (ret == 1) {  // connected

      log_printf(LOG_DEBUG, "[%s] Connected to passive port\n", node->name);
      node->state.substate++;
      break;

    }


    if (ret == 2) {  // failed

      log_printf(LOG_ERROR, "[%s] Failed to connect to passive\n", node->name);
      node->state.substate = 999;    // error
      break;

    }

    break;



  case 83:
    //  Connected, send list command.
    
    site_printf(node, "LIST -l\r\n");
    node->state.substate++;
    break;


  case 84:
    // send ls, wait for 150 start
    if (site_waitfor(node,  150)) { // We've started
      node->state.substate++;
    }
    break;


  case 85:
    // wait for 226 finish line
    if (site_waitfor(node,  226)) { // We've started
      node->state.substate++;
    }
    break;

  case 86:
    // Wait for data session to close.

    if (node->state.pasv == SO_NONE) {

      sockets_close(node->pasv_socket);
      node->pasv_socket = 0;

      // We are done collecting
      // FIXME
      node->state.substate++;

    }

    if (node->state.pasv == SO_ERROR) {

      sockets_close(node->pasv_socket);
      node->pasv_socket = 0;

      // Failed to get directory
      node->state.substate = 999;    // error

    }
    break;




  case 87:
    log_printf(LOG_DEBUG, "[%s] Retrieved directory\n", node->name);

    node->state.substate++;
    break;


  case 88:
    if ((mirror->A_host.state.substate == 88) &&
	((mirror->B_host.state.substate == 88) ||
	 ((mirror->B_host.state.substate == 999)))) {

      mirror->B_host.state.substate = 0;
      mirror->A_host.state.substate++;
      break;

    }

    break;

  case 89:
    log_printf(LOG_DEBUG, "Both directories retrieved - comparing\n");

    // Work out which entries to move/copy.
    compare_dirs(mirror, &(mirror->A_host.mirror_current->subdir));


    if (!mirror->A_host.mirror_current->subdir) {
      // Go back up a directory

      log_printf(LOG_DEBUG, "No work to be done here..\n");

      site_waitclear(&mirror->A_host);
      site_waitclear(&mirror->B_host);

      site_printf(&(mirror->A_host), "CWD ..\r\n");
      site_printf(&(mirror->B_host), "CWD ..\r\n");

      mirror->A_host.state.substate = 96;


      if (mirror->move_empty_dirs == YN_YES)
	mirror->B_host.state.substate = 96;
      else
	mirror->B_host.state.substate = 94;


      break;


    } else {


      log_printf(LOG_DEBUG, "Entering subdir\n");


    }

    mirror->A_host.state.substate = 0;
    mirror->B_host.state.substate = 0;
    break;


  case 94:
    if (site_waitfor(node, 250))
      node->state.substate++;
    break;


  case 95:
    site_printf(&mirror->B_host, "RMD %s\r\n", 
		mirror->A_host.mirror_current->name);
    node->state.substate = 97;
    break;


  case 96:
    if (site_waitfor(node, 250))
      node->state.substate++;
    break;


  case 97:
    if ((mirror->A_host.state.substate == 97) &&
	(mirror->B_host.state.substate == 97)) {
      mirror->A_host.state.substate = 0;
      mirror->B_host.state.substate = 0;
      break;
    }
    break;

  case 98:
    if (site_waitfor(node, 250))
      node->state.substate++;
    break;


  case 99:
    if ((mirror->A_host.state.substate == 99) &&
	(mirror->B_host.state.substate == 99)) {
      mirror->A_host.state.substate = 1;
      mirror->B_host.state.substate = 1;
      break;
    }
    break;


  case 999:
      log_printf(LOG_ERROR, "%s (failed -- skipping entry)\n", mirror->A_host.mirror_current->name);
      mirror->A_host.state.substate = 0;
      mirror->B_host.state.substate = 0;

      node->stats.skipped_files++;

      break;


  default:
    log_printf(LOG_ERROR, "Reached unknown substate %d -- int(f)ernal bug.\n",
	   mirror->A_host.state.substate);
    break;

  }


}



//
// Advance the mirror queue. Checking subdirs, parent dirs etc.
//
void state_advance(mirror_t *mirror)
{
  static int skip = 0;

  if (skip) {
    skip = 0;
    return;
  }


  log_printf(LOG_DEBUG, "Advancing queue\n");


  // First time
  if (!mirror->A_host.mirror_current) {

    log_printf(LOG_DEBUG, "Initial\n");

    mirror->A_host.mirror_current = mirror->A_host.mirror_paths;
    return;

  }


  mirror->A_host.retries_done = 0;
  mirror->B_host.retries_done = 0;



  if (mirror->A_host.mirror_current->subdir) {

    log_printf(LOG_DEBUG, "Going lower\n");

    mirror->A_host.mirror_current = mirror->A_host.mirror_current->subdir;
    return;

  }



  //
  // Try to obtain either a ->next or a ->parent's ->next.
  // 

  while (1) { 
    
    
    // Check if there is a next
    if (mirror->A_host.mirror_current->next) {
      

      if (!mirror->A_host.mirror_current->parent) {

	// We have no parent, then we remember this date.
	// If next node has same date, skip this.

	if (mirror->A_host.mirror_current->next->date !=
	    mirror->A_host.mirror_current->date) {

	  date_stamp( mirror, &mirror->A_host, mirror->A_host.mirror_current);



	}
      }


      log_printf(LOG_DEBUG, "Following next\n");

      mirror->A_host.mirror_current = mirror->A_host.mirror_current->next;
      return;
      
    }
    
    // If there is no next, check if we have parent
    
    if (mirror->A_host.mirror_current->parent) {
      
      log_printf(LOG_DEBUG, "Going up\n");
      log_printf(LOG_MIRROR, "%s (finished directory)\n", mirror->A_host.mirror_current->name);


      mirror->A_host.mirror_current = mirror->A_host.mirror_current->parent;
    
      // Do "CWD .."
      site_waitclear(&mirror->A_host);
      site_waitclear(&mirror->B_host);

      site_printf(&(mirror->A_host), "CWD ..\r\n");
      site_printf(&(mirror->B_host), "CWD ..\r\n");

      mirror->A_host.state.substate = 96;
      mirror->B_host.state.substate = 96;

      skip = 1;

      continue;
      
    }
    
    
    
    
    // No parent and no next, then the queue must be empty. We are done.
    if (!mirror->A_host.mirror_current->parent) {

      date_stamp( mirror, &mirror->A_host, mirror->A_host.mirror_current);

      log_printf(LOG_DEBUG, "Queue is empty, work complete\n");

      log_printf(LOG_DEBUG, "Advancing to next mirror path\n");

      pasv_freedir(&(mirror->A_host.mirror_paths));
      pasv_freedir(&(mirror->B_host.mirror_paths));

      mirror->A_host.mirror_current = NULL;
      mirror->B_host.mirror_current = NULL;

      mirror->A_host.retries_done = 0;
      mirror->B_host.retries_done = 0;

      collect_setpaths( mirror );

      return;
      
    }
    
  } // while


}





void state_done(mirror_t *mirror, site_t *node)
{

  site_printf(node, "QUIT\r\n");
  site_disconnect(node);

  if ((mirror->A_host.state.socket == SO_NONE) &&
      (mirror->B_host.state.socket == SO_NONE))
    io_setexit();

}



//
// One or more site's failed at some point, sleep and retry.
//
void state_retry(mirror_t *mirror, site_t *node)
{


  switch(node->state.substate) {

  case 0:

    if (mirror->A_host.state.state != ST_RETRY)
      state_set(&mirror->A_host, ST_RETRY);

    if (mirror->B_host.state.state != ST_RETRY)
      state_set(&mirror->B_host, ST_RETRY);


    if ((node->state.socket != SO_NONE) && (node->socket >= 0)) {
      
      sockets_close(node->socket);
      node->state.socket = SO_NONE;
      node->socket = -1;
      
      log_printf(LOG_DEBUG, "[%s] socket closed\n", node->name);
      
    }

    if ((node->state.pasv != SO_NONE) && 
	(node->pasv_socket > 0)) {

      sockets_close(node->pasv_socket);
      node->pasv_socket = 0;
      node->state.pasv = SO_NONE;

      log_printf(LOG_DEBUG, "[%s] data socket closed\n", node->name);

    }


    
    log_printf(LOG_DEBUG, "[%s] sleeping for retries\n", node->name);


    node->idle = io_global_time;


    node->state.substate++;
    break;


  case 1:
    // Have we retries enough times?

    if (node->retries_done >= node->retries)
      node->state.substate = 20;

    // Have we waited retries seconds?
    if ((io_global_time - node->idle) > node->retries) {

      node->retries_done++;

      state_reset(node);

      //      io_setexit();


      break;
      
    }

    break;

  case 20:
    // Sleep instead
    log_printf(LOG_DEBUG, "Deep sleep\n");
    node->state.substate++;
    break;

  case 21:

    if ((io_global_time - node->idle) > node->sleep) {

      node->retries_done++;

      state_reset(node);

      //io_setexit();

      break;
    }

    break;

  }


}


